layouts = {
    "grid": {
        "units": "mm",
        "LaTeX_font_family": "libertine",
        "background_color": [ 255, 255, 255 ],

        "row_space": 0.8,
        "column_space": 0.8,

        "padding": {
            "north": 0.0,
            "west": 0.0,
            "east": 0.0,
            "south": 0.0
        },

        "element_config": {
            "captions": {
                "north": {
                    "height": 0.0,
                    "offset": 0.0,
                    "rotation": 0,
                    "fontsize": 6,
                    "line_space": 1.2,
                    "text_color": [ 0, 0, 0 ]
                },

                "south": {
                    "height": 0.0,
                    "offset": 0.0,
                    "rotation": 0,
                    "fontsize": 6,
                    "line_space": 1.2,
                    "text_color": [ 0, 0, 0 ]
                },

                "east": {
                    "width": 0.0,
                    "offset": 0.0,
                    "rotation": -90,
                    "fontsize": 6,
                    "line_space": 1.2,
                    "text_color": [ 0, 0, 0 ]
                },

                "west": {
                    "width": 0.0,
                    "offset": 0.0,
                    "rotation": 90,
                    "fontsize": 6,
                    "line_space": 1.2,
                    "text_color": [ 0, 0, 0 ]
                }
            }
        },
        "titles": {
            "north": {
                "height": 0.0,
                "offset": 0.0,
                "rotation": 0,
                "fontsize": 8,
                "line_space": 1.2,
                "text_color": [ 0, 0, 0 ],
                "background_color": None
            },
            "south": {
                "height": 0.0,
                "offset": 0.0,
                "rotation": 0,
                "fontsize": 8,
                "line_space": 1.2,
                "text_color": [ 0, 0, 0 ],
                "background_color": None
            },
            "west": {
                "width": 0.0,
                "offset": 0.0,
                "rotation": 90,
                "fontsize": 8,
                "line_space": 1.2,
                "text_color": [ 0, 0, 0 ],
                "background_color": None
            },
            "east": {
                "width": 0.0,
                "offset": 0.0,
                "rotation": -90,
                "fontsize": 8,
                "line_space": 1.2,
                "text_color": [ 0, 0, 0 ],
                "background_color": None
            }
        },
        "row_titles": {
            "west": {
                "width": 0.0,
                "offset": 0.0,
                "rotation": 90,
                "fontsize": 7,
                "line_space": 1.2,
                "text_color": [ 0, 0, 0 ],
                "background_colors": None
            },
            "east": {
                "width": 0.0,
                "offset": 0.0,
                "rotation": -90,
                "fontsize": 7,
                "line_space": 1.2,
                "text_color": [ 0, 0, 0 ],
                "background_colors": None
            }
        },
        "column_titles": {
            "north": {
                "height": 0.0,
                "offset": 0.0,
                "rotation": 0,
                "fontsize": 7,
                "line_space": 1.2,
                "text_color": [ 0, 0, 0 ],
                "background_colors": None
            },
            "south": {
                "height": 0.0,
                "offset": 0.0,
                "rotation": 0,
                "fontsize": 7,
                "line_space": 1.2,
                "text_color": [ 0, 0, 0 ],
                "background_colors": None
            }
        }
    }
}